/* urmpconfig.c -- 
 * Created: Tue Dec 30 10:30:44 2008 by faith@acm.org
 * Revised: Sun Apr 12 12:16:24 2009 by faith@acm.org
 * Copyright 2008, 2009 Rickard E. Faith (faith@acm.org)
 * This program comes with ABSOLUTELY NO WARRANTY.
 * 
 * $Id$
 * 
 */

#include "urmpconfig.h"
#include "urmpgui.h"
#include <wordexp.h>
#include <time.h>
#include <sys/stat.h>

static bool urmp_config_is_readonly;
const char  *urmp_config_dir_path;
const char  *urmp_config_db_path;
const char  *urmp_config_backup_db_path;
const char  *urmp_config_file_path;
ur_parse_t  urmp_parse_use;
ur_parse_t  urmp_parse_augment;
ur_parse_t  urmp_parse_priority;

static struct {
	int width;
	int height;
	int division;
	int widths[URMP_PANES][URMP_COLUMNS];
} urmp_config;

typedef	enum {
	state_normal,
	state_use,
	state_augment
} urmp_config_state_t;

void urmp_config_readonly(void)
{
	urmp_config_is_readonly = true;
}

static void urmp_config_filter_dump(FILE *str, const char *name,
				    bool priority, ur_parse_t p)
{
	int i;
	int filter;
	const char *items[] = { "title", "artist", "album", "track" };

	if (!p) return;
	if (priority) {
		for (filter = 0; ur_parse_get_filter(p, filter); filter++) {
			fprintf(str, "%s %d %s %s\n",
				name, ur_parse_get_attribute_int(p, filter),
				ur_parse_get_attribute_string(p, filter),
				ur_parse_get_filter(p, filter));
		}
		return;
	}

	for (filter = 0; ur_parse_get_filter(p, filter); filter++) {
		fprintf(str, "%s %s\n", name, ur_parse_get_filter(p, filter));
		for (i = 0; i < 4; i++) {
			const char *pattern = ur_parse_get_item(p, filter, i);
			if (!pattern) continue;
			fprintf(str, "%s %s\n", items[i], pattern);
		}
	}
}

static void urmp_config_write(const char *filename)
{
	FILE   *str;
	time_t t;
	int    i, j;

	if (urmp_config_is_readonly) return;

	if (!(str = fopen(filename, "w"))) {
		ur_fatal_errno(__func__,
			       "Cannot open %s for write\n", filename);
	}
	time(&t);
	fprintf(str, "# Generated by urmp %24.24s\n", ctime(&t));
	fprintf(str, "width %d\n", urmp_config.width);
	fprintf(str, "height %d\n", urmp_config.height);
	fprintf(str, "division %d\n", urmp_config.division);
	for (i = 0; i < URMP_PANES; i++) {
		for (j = 0; j < URMP_COLUMNS; j++) {
			fprintf(str, "column %d %d %d\n", i, j,
				urmp_config.widths[i][j]);
		}
	}
	urmp_config_filter_dump(str, "augment",  false, urmp_parse_augment);
	urmp_config_filter_dump(str, "use",      false, urmp_parse_use);
	urmp_config_filter_dump(str, "priority", true,  urmp_parse_priority);
	fclose(str);
}

static void urmp_config_set_defaults(void)
{
	int i, j;

	urmp_config.width    = URMP_DEFAULT_WIDTH;
	urmp_config.height   = URMP_DEFAULT_HEIGHT;
	urmp_config.division = URMP_DEFAULT_WIDTH / 2;

	for (i = 0; i < URMP_PANES; i++) {
		for (j = 0; j < URMP_COLUMNS; j++) {
			urmp_config.widths[i][j]
				= URMP_DEFAULT_WIDTH/2/(URMP_COLUMNS+1);
		}
	}
}

static int urmp_config_item_add(const char *field,
				urmp_config_state_t state,
				int item,
				int target,
				const char *pattern)
{
	if (item > target) {
		ur_internal(__func__,
			    "Field \"%s\" unexpected here, state=%d, item=%d,"
			    " target=%d, pattern=%s\n",
			    field, state, item, target, pattern);
	}

	switch (state) {
	case state_use:
		for (; item < target; item++) {
			ur_parse_item_add(urmp_parse_use, NULL);
		}
		ur_parse_item_add(urmp_parse_use, pattern);
		break;
	case state_augment:
		for (; item < target; item++) {
			ur_parse_item_add(urmp_parse_augment, NULL);
		}
		ur_parse_item_add(urmp_parse_augment, pattern);
		break;
	default:
		ur_internal(__func__,
			    "Field \"%s\" unexpected here, state=%d, item=%d,"
			    " target=%d, pattern=%s\n",
			    field, state, item, target, pattern);
	}
	return ++item;
}

static void urmp_config_read(const char *filename)
{
	FILE                *str;
	char                buf[1024];
	char                *field;
	char                *value;
	ur_list_t           list;
	int                 item = 0;
	urmp_config_state_t state = state_normal;

	if (!(str = fopen(filename, "r"))) return;
	while (fgets(buf, sizeof(buf)-1, str)) {
		if (buf[0] == '#' || buf[0] == '!') continue;

		list = ur_list_create();
		ur_list_parse(list, buf, UR_LIST_DELIM, UR_LIST_QUOTE,
			      NULL, false);
		field = ur_list_pop(list);
		value = ur_list_pop(list);
		if (!field || !value) {
			ur_list_destroy(list);
			continue;
		}
		if (!strcmp(field, "width")) {
			urmp_config.width = strtol(value, NULL, 0);
			state = state_normal;
		}
		if (!strcmp(field, "height")) {
			urmp_config.height = strtol(value, NULL, 0);
			state = state_normal;
		}
		if (!strcmp(field, "division")) {
			urmp_config.division = strtol(value, NULL, 0);
			state = state_normal;
		}
		if (!strcmp(field, "column")) {
			int pane   = strtol(value, NULL, 0);
			int column = strtol(ur_list_pop(list), NULL, 0);
			int v      = strtol(ur_list_pop(list), NULL, 0);
			if (pane < 0
			    || pane > URMP_PANES
			    || column < 0
			    || column > URMP_COLUMNS) {
				ur_internal(__func__,
					    "pane=%d column=%d width=%d\n",
					    pane, column, v);
			}
			urmp_config.widths[pane][column] = v;
			state = state_normal;
		}
		/*
		 * use <pattern>
		 * augment <pattern>
		 * priority <priority> <color> <pattern>
		 *
		 * title, artist, album, track
		 */
		if (!strcmp(field, "use")) {
			if (!urmp_parse_use) {
				urmp_parse_use = ur_parse_create();
			}
			ur_parse_filter_add(urmp_parse_use, value);
			state = state_use;
			item  = 0;
		}
		if (!strcmp(field, "augment")) {
			if (!urmp_parse_augment) {
				urmp_parse_augment = ur_parse_create();
			}
			ur_parse_filter_add(urmp_parse_augment, value);
			state = state_augment;
			item  = 0;
		}
		if (!strcmp(field, "priority")) {
			int        priority = strtol(value, NULL, 0);
			const char *color   = ur_list_pop(list);

			if (!urmp_parse_priority) {
				urmp_parse_priority = ur_parse_create();
			}
			ur_parse_filter_add(urmp_parse_priority,
					    ur_list_pop(list));
			ur_parse_attribute_int(urmp_parse_priority, priority);
			ur_parse_attribute_string(urmp_parse_priority, color);
			state = state_normal;
		}
		if (!strcmp(field, "title")) {
			item = urmp_config_item_add(field, state, item,
						    0, value);
		}
		if (!strcmp(field, "artist")) {
			item = urmp_config_item_add(field, state, item,
						    1, value);
		}
		if (!strcmp(field, "album")) {
			item = urmp_config_item_add(field, state, item,
						    2, value);
		}
		if (!strcmp(field, "track")) {
			item = urmp_config_item_add(field, state, item,
						    3, value);
		}
		ur_list_destroy(list);
	}
	fclose(str);

	if (!urmp_parse_augment) {
		urmp_parse_augment = ur_parse_create();
		ur_parse_filter_add(urmp_parse_augment, ".*/\\d+ - .*.mp3");
		ur_parse_item_add(urmp_parse_augment, ".*/\\d+ - (.*).mp3");
		ur_parse_item_add(urmp_parse_augment,
				  ".*/(.*)/.*/\\d+ - .*.mp3");
		ur_parse_item_add(urmp_parse_augment, ".*/(.*)/\\d+ - .*.mp3");
		ur_parse_item_add(urmp_parse_augment, ".*/(\\d+) - .*.mp3");

		ur_parse_filter_add(urmp_parse_augment, "/[^/-]*-[^/-]*.mp3");
		ur_parse_item_add(urmp_parse_augment, "/[^/-]*-([^/-]*).mp3");
		ur_parse_item_add(urmp_parse_augment, "/([^/-]*)-[^/-]*.mp3");

		ur_parse_filter_add(urmp_parse_augment, ".*.mp3");
		ur_parse_item_add(urmp_parse_augment, ".*/.*/(.*).mp3");
		ur_parse_item_add(urmp_parse_augment, ".*/(.*)/.*.mp3");

		urmp_config_write(filename);
	}
}

void urmp_config_set_column_width(int pane, int column, int width)
{
	if (pane < 0
	    || pane > 1
	    || column < 0
	    || column > URMP_COLUMNS) {
		ur_internal(__func__, "pane=%d column=%d width=%d\n",
			    pane, column, width);
	}
	if (urmp_config.widths[pane][column] == width) return;
	urmp_config.widths[pane][column] = width;
	urmp_config_write(urmp_config_file_path);
}

int urmp_config_get_column_width(int pane, int column)
{
	if (pane < 0
	    || pane > URMP_PANES
	    || column < 0
	    || column > URMP_COLUMNS) {
		ur_internal(__func__, "pane=%d column=%d\n",
			    pane, column);
	}
	return urmp_config.widths[pane][column];
}

void urmp_config_set_division(int division)
{
	if (urmp_config.division == division) return;
	urmp_config.division = division;
	urmp_config_write(urmp_config_file_path);
}

int urmp_config_get_division(void)
{
	return urmp_config.division;
}

void urmp_config_set_geometry(int width, int height)
{
	bool changed = false;

	if (urmp_config.width != width
	    || urmp_config.height != height) {
		changed = true;
	}
	if (!changed) return;

	urmp_config.width = width;
	urmp_config.height = height;
	urmp_config_write(urmp_config_file_path);
}

int urmp_config_get_width(void)
{
	return urmp_config.width;
}

int urmp_config_get_height(void)
{
	return urmp_config.height;
}

void urmp_config_setup(const char *config_path, bool use_defaults)
{
	const char  *config;
	wordexp_t   p;
	struct stat st;
	int         len;
	char        path[urmp_max_pathlen+1];
	char        backup_path[urmp_max_pathlen+1];

	if (config_path) {
		config = config_path;
	} else if (!(config = getenv(URMP_ENV_DEFAULT_CONFIG))) {
		config = URMP_DEFAULT_CONFIG;
	}

	ur_trace(URMP_TRACE_CONFIG, __func__,
		 "Config directory: %s\n", config);

	len = strlen(config);

	wordexp(config, &p, 0);
	if (!p.we_wordv)
		ur_fatal(__func__,
			 "Invalid configuration specification: %s\n", config);
	urmp_config_dir_path = ur_atom_string(p.we_wordv[0]);
	wordfree(&p);

	if (stat(urmp_config_dir_path, &st)) {
		if (mkdir(urmp_config_dir_path, 0755))
			ur_fatal_errno(__func__, "Cannot create %s\n",
				       urmp_config_dir_path);
	} else if (!S_ISDIR(st.st_mode)) {
		ur_fatal(__func__,
			 "Configuration directory not a directory: %s\n",
			 urmp_config_dir_path);
	}

	if (!urmp_config_db_path) {
		snprintf(path, urmp_max_pathlen,
			 "%s/%s", config, URMP_DEFAULT_DB);
	}
	wordexp(path, &p, 0);
	if (!p.we_wordc)
		ur_fatal(__func__, "Invalid db path: %s\n", path);
	urmp_config_db_path = ur_atom_string(p.we_wordv[0]);

	snprintf(backup_path, urmp_max_pathlen, "%s.bak", p.we_wordv[0]);
	urmp_config_backup_db_path = ur_atom_string(backup_path);
	wordfree(&p);

	if (!urmp_config_file_path) {
		snprintf(path, urmp_max_pathlen,
			 "%s/%s", config, URMP_DEFAULT_RC);
	}
	wordexp(path, &p, 0);
	if (!p.we_wordc)
		ur_fatal(__func__, "Invalid config file path: %s\n", path);
	urmp_config_file_path = ur_atom_string(p.we_wordv[0]);

	urmp_config_set_defaults();
	if (!use_defaults) urmp_config_read(urmp_config_file_path);
}
